UNIT FileHash;
{$IFDEF FPK}
{$PACKRECORDS 1}
{$ENDIF}
(*
 * Universale File-Hashing-Unit fr STRING-Keys.
 * (c) 1995 Hans-Peter Zorn  <hzorn@ix.urz.uni-heidelberg.de>
 * Version 0.4
 *
 * Diese UNIT ist frei fr jede nicht-kommerzielle Nutzung, ber
 * eine Erwhnung in der Doku von damit erstellten Programmen
 * wrde ich mich freuen!
 *
 * nderungen im Source bitte mit Namen und Adresse markieren,
 * wenn er weitergegeben wird. Bitte ein Kopie an mich schicken.
 *
 * ChangeLog:
 * 14.11.95: Alles :-)
 * 15.11.95: Felder mit key='' sind gelscht definiert.
 *           GetData und InsertData entsprechend gendert.
 *           SetData und Delete ergnzt.
 * 20.01.96: Der Stream ist nun nichtmehr permananent geffnet,
 *           der Dateiname wird in einem dynamischen String
 *           "Dateiname" gespeichert.
 * 25.12.97  Anpassungen an FPK-Pascal
 *
 *)

INTERFACE

USES OBJECTS;

CONST

(*
 * Maximale Lnge eines Schlssels
 *)


   KeyLength=20;


TYPE


  KeyString=STRING[KeyLength];
   ByteArray=ARRAY[0..60000] OF BYTE;
  (*
   * RECORD THashEntry
   *
   * Dieser RECORD ist das, was auf die Platte geschrieben wird.
   * THashEntry.Key  ist der Schlssel
   * THashEntry.Next ist die Position im File, wo der nchste
   *                 RECORD mit dem selben Hash-Wert, wenn
   *                 eine Kollision aufgetreten ist.
   *                 Wenn kein weiterer Record existiert, ist
   *                 dieses Feld leer
   * THashEntry.Data Dieses Feld enthlt die Applikations-Spezifischen
   *                 Daten. Die Grsse wird bei TFHash.Create angegeben
   *                 und danach in den Bytes 5-6 des Hashfiles gespeichert
   *)

  THashEntry=RECORD
     key:KeyString;
     next:longint;
     data:ByteArray;
  END;

  (*
   *  OBJECT TFHash
   *
   *  Dieses Object enthlt eigentlich das gesamte Hashing-Modul.
   *
   *  TFHash.HFile    dies ist ein Object vom Typ TSTREAM aus der
   *                  TP Unit OBJECTS. Die Standard-Konstruktoren
   *                  dieses Moduls verwenden den TYP TDosStream,
   *                  der ganz einfaches DOS File-I/O verwendet.
   *                  Durch berladen des Konstructors kann man aber
   *                  hier auch TBufStream,TEMSStream oder TXMSStream
   *                  einsetzen.
   *
   *  TFHash.N        Dies ist die Maximalanzahl der Hasheintrge.
   *                  Wenn als Hash-Funktion h(k)=k MOD N eingesetzt wird,
   *                  wie von dem Object als default, so ist hier mglichst
   *                  eine Primzahl zu verwenden. Der Wert wird bei
   *                  TFHash.Create angegeben und dann in den Bytes
   *                  1 bis 4 des Hashfiles gespeichert
   *
   *  TFHash.ActEntry Wird nur intern verwendet um die Hasheintrge
   *                  aus dem File zu lesen.
   *
   *  TFHash.EntrySize
   *                  Grsse eines Hasheintrages, berechnet sich aus
   *                  SizeOf(keyString)+SizeOf(longint)+"Grsse von DATA"
   *
   *  TFHash.Init     Dieser Constructor ffnet ein bereits vorhandenes
   *                  Hashfile, liest die Werte fr N und EntrySize und
   *                  alloziert ActEntry.
   *   * Parameter:
   *                  datei                  Dateiname des Hashfiles
   *
   *  TFHash.Create   Dieser Constructor erzeugt ein neues Hashfile.
   *                  Ansonsten wir Init.
   *                  VORSICHT: ein vorhandenes File wird berschrieben!
   *   * Parameter:
   *                  datei                  Dateiname des Hashfiles
   *                  size                   Wert fr N, also anzahl
   *                                         der Hasheintrge, sollte
   *                                         ca die zu erwartende Anzahl
   *                                         mal 2 sein.
   *                  datasize               Die grsse der Daten, die
   *                                         die Applikation unter einem
   *                                         Key abspeichern will.
   *
   *  TFHash.Done     Destructor, schliesst das File und gibt den
   *                  Speicher frei.
   *
   *  TFHash.GetData  Sucht nach einem Key im Hashfile und gibt
   *                  Erfolg/kein Erfolg zurck. Ausserdem werden
   *                  bei Erfolg die zum key gehrigen Daten
   *                  zurckgeliefert.
   *   * Parameter:
   *                  Key                    der zu suchende Schlssel
   *                  Data                   Puffer, der bei Erfolg mit
   *                                         den Daten gefllt wird.
   *                                         MUSS GENGEND GROSS SEIN!
   *   * Rckgabe:
   *                  BOOLEAN: TRUE wenn gefunden, FALSE wenn nicht.
   *
   *  TFHash.InsertData
   *                  Fgt den Schlssel und die Daten in das Hashfile ein.
   *                  Wenn der Schlssel bereits vorhanden ist, werden die
   *                  Daten im File *nicht* gendert.
   *   * Parameter:
   *                  Key                    der Schlssel
   *                  Data                   Puffer mit den zu speichernden
   *                                         Daten.
   *   * Rckgabe:    BYTE:
   *                     0         Key war schon vorhanden, nichts
   *                               getan.
   *                     1         Key und Daten wurden an erster Position
   *                               eingefgt.
   *                    >2         Wie 1, wert ist Anzahl der Verkettungen
   *                               wegen Kollision plus 2
   *
   *  TFHash.SetData
   *                  Wie InsertData, nur werden hier die Daten neu geschrieben,
   *                  wenn der Schlssel vorhanden war.
   *
   *  TFHash.Delete   Markiert einen Eintrag als gelscht, wirkliches Lschen
   *                  ist leider nicht mglich.
   *
   *         SetData und Delete geben TRUE/FALSE fr key gefunden/nicht gefunden
   *         zurck.
   *
   *  TFHash.KondStr  Wandelt einen String in einen LONGINT um. Die originale
   *                  Funktion verwendet den im Buch "Algorithmen" von R Sedgewick
   *                  beschriebenen Algorithmus.
   *                  Sie kann je nach Anforderungen der Applikation berladen
   *                  werden
   *
   *  TFHash.HashFunc Bildet einen bergebenen LONGINT auf einen Wert innerhalb
   *                  des Bereiches von 0 bis N ab. Wenn nicht berladen, so
   *                  ist diese Funktion H(k) = k MOD N.
   *)

  PFHASH=^TFHASH;
  TFHASH=OBJECT
    HFile          : PStream;
    N              : LONGINT;
    ActEntry       : ^THashEntry;
    Entrysize      : WORD;
    dateiname      : ^STRING;
    CONSTRUCTOR   init(datei:string);
    CONSTRUCTOR   create(datei:STRING;size:LONGINT;DataSize:WORD);
    DESTRUCTOR    done;
    FUNCTION      GetData(key : keySTRING; data:pointer):BOOLEAN;
    FUNCTION      SetData(key:keySTRING;var data):BOOLEAN;
    FUNCTION      Delete(key:keySTRING;var data):BOOLEAN;
    FUNCTION      InsertData(key:KeyString;var DATA):BYTE;
    FUNCTION      KondStr(key:KeySTRING):LONGINT;
    FUNCTION      HashFunc(k:LONGINT):LONGINT;
  END;

  (*
   * Diese Variable wird <>0 wenn Fehler aufgetreten sind, unbedingt
   * nach dem Aufruf der Constructors abfragen!
   *)

CONST
  HashError:BYTE=0;

IMPLEMENTATION
{$I-}
CONST

(*
 * Anzahl der nicht zur Hash-Tabelle gehrigen Bytes am Anfang des
 * Files. In diesem Bereich stehen File-ID, die werte fr N und EntrySize
 *
 *)

   FileOffset=16;

(*
 * String am Anfang des Files, an dem erkannt wird, ob es sich
 * um ein Hashing-File handelt.
 *
 *)

   FileID:STRING[9]='HPZH03'#13#10#26;
   ControlSize:BYTE=sizeof(KeyString)+SizeOf(longint);

PROCEDURE move(VAR a,b;size : LONGINT);
VAR W:LONGINT;
BEGIN
   FOR W:=0 TO SIZE-1 DO
      ByteArray(b)[W]:=ByteArray(a)[W];
END;				    

CONSTRUCTOR TFHASH.create(datei:STRING;size:LONGINT;DataSize:WORD);
VAR w:WORD;b:BYTE;
BEGIN
  HashError:=0;
  {Stream initialisieren}
  GetMem(dateiname,Length(Datei)+1);
  HFile:=New(PDosStream,Init(datei,stcreate));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=2;
    Fail;
  END;
  N:=size;
  HFile^.Write(FileID,10);
  HFile^.Write(N,4);
  EntrySize:=SizeOf(keyString)+SizeOf(LongInt)+DataSize;
   
  HFile^.Write(EntrySize,2);
  GetMem(ActEntry,EntrySize);
  {Leeres HashFile erzeugen}
  Fillchar(ActEntry^,EntrySize,0);
  FOR w:=0 to N DO
    HFile^.Write(actEntry^,EntrySize);
  IF HFile^.Status<>stOK THEN
  BEGIN
    Freemem(actentry,entrysize);
    dispose(HFile,done);
    HashError:=3;
    Fail;
  END;
  dispose(hfile,done);
  GetMem(dateiname,Length(Datei)+1);
  DateiName^:=Datei;
END;


CONSTRUCTOR TFHASH.init(datei:string);
var S:STRING[9];B:BYTE;
BEGIN
  {Stream initialisieren}
  HashError:=0;
  HFile:=New(PDosStream,Init(datei,stOpen));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=1;
    Fail;
  END;
  HFile^.Read(s,10);
  IF s<>FileId THEN
  BEGIN
    dispose(hfile,done);
    HashError:=4;
    Fail;
  END;
  HFile^.Read(N,4);
  HFILE^.Read(EntrySize,2);
  dispose(hfile,done);
  GetMem(ActEntry,EntrySize);
  GetMem(dateiname,Length(Datei)+1);
  DateiName^:=Datei;
END;

DESTRUCTOR TFHash.Done;
VAR B:BYTE;
BEGIN
  freemem(ActEntry,EntrySize);
  freemem(dateiname,length(Dateiname^)+1);
END;

FUNCTION TFHASH.KondStr(key:KeySTRING):LONGINT;
VAR l:longint;b,size:byte;bits,maxbytes:BYTE;
BEGIN
  l:=0;
  SIZE:=Length(Key);
  l:=ord(key[1]);
  FOR B:=2 TO SIZE DO
    l:= ((l*48)+ord(key[b])) MOD N;
  KondStr:=l;
END;

FUNCTION TFHASH.HashFunc(k:longint):LONGINT;
BEGIN
  HashFunc:=abs(k mod N);
END;


FUNCTION TFHASH.GetData(key : keySTRING;data: pointer):BOOLEAN;
VAR p:LONGINT;
BEGIN
   fillchar(actentry^,entrysize,0);
  HFile:=New(PDosStream,Init(dateiname^,stOpen));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=1;
    exit;
  END;
  HashError:=0;
  p:=HashFunc(KondStr(key));
  p:=FileOffset+(p*entrysize);
  HFile^.Seek(p);
  HFile^.Read(ActEntry^,EntrySize);
  WHILE (ActEntry^.Key<>Key) AND (ActEntry^.next<>0) DO
  BEGIN
    HFile^.Seek(actEntry^.Next);
    HFile^.Read(actEntry^,EntrySize);
  END;
  IF ActEntry^.key<>key THEN
    GetData:=FALSE
  ELSE
  BEGIN
{     WriteLn(entrysize,' ',actentry^.key,' ');
     readln;}
     fillchar(data^,0,4);
    move(ActEntry^.data[0],bytearray(data^)[0],Entrysize-ControlSize);
   {  longint(data^):=1;}
    GetData:=TRUE;
  END;
  dispose(hfile,done);
END;

FUNCTION TFHASH.InsertData(key:keySTRING;var data):BYTE;
VAR p,p2:LONGINT;collcount:byte;
BEGIN
  HFile:=New(PDosStream,Init(dateiname^,stOpen));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=1;
    exit;
  END;
  HashERror:=0;
  collcount:=0;
  p:=HashFunc(KondStr(key));
  p:=FileOffset+(p*EntrySize);
  p2:=p;
  HFile^.Seek(p);
  HFile^.Read(ActEntry^,EntrySize);
  WHILE (ActEntry^.Key<>'') AND (ActEntry^.Key<>key) AND (ActEntry^.next<>0) DO
  BEGIN
    inc(collcount);
    p2:=actEntry^.Next;
    HFile^.Seek(actEntry^.Next);
    HFile^.Read(actEntry^,EntrySize);
  END;
  IF ActEntry^.key=key THEN
    InsertData:=0
  ELSE
  BEGIN
    {Position ist unbesetzt -> eintragen!}
    IF (actentry^.key='') THEN
    BEGIN
      ActEntry^.Key:=Key;
      move(bytearray(data),ActEntry^.Data,EntrySize-ControlSize);
      HFile^.Seek(p);
      HFile^.Write(ActEntry^,EntrySize);
      InsertData:=1;
    END
    ELSE
    {Kollision -> in die Liste einfgen}
    BEGIN
      inc(collcount);
      p:=HFile^.GetSize;
      ActEntry^.Next:=p;
      HFile^.Seek(p2);
      HFile^.Write(ActEntry^,EntrySize);
      ActEntry^.Key:=Key;
      ActEntry^.Next:=0;
      move(bytearray(data),ActEntry^.Data,EntrySize-ControlSize);
      HFile^.Seek(p);
      HFile^.Write(ActEntry^,EntrySize);
      InsertData:=collcount+2;
    END;
  END;
  dispose(hfile,done);
END;

FUNCTION TFHASH.SetData(key:keySTRING;var data):BOOLEAN;
VAR p,p2:LONGINT;collcount:byte;
BEGIN
  HFile:=New(PDosStream,Init(dateiname^,stOpen));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=1;
    exit;
  END;
  collcount:=0;
  p:=HashFunc(KondStr(key));
  p:=FileOffset+(p*EntrySize);
  p2:=p;
  HFile^.Seek(p);
  HFile^.Read(ActEntry^,EntrySize);
  WHILE (ActEntry^.Key<>'') AND (ActEntry^.Key<>key) AND (ActEntry^.next<>0) DO
  BEGIN
    inc(collcount);
    p2:=actEntry^.Next;
    HFile^.Seek(actEntry^.Next);
    HFile^.Read(actEntry^,EntrySize);
  END;
  IF ActEntry^.key=key THEN
  BEGIN
    SetData:=TRUE;
    move(bytearray(data),ActEntry^.Data,EntrySize-ControlSize);
    HFile^.Seek(p2);
    HFile^.Write(ActEntry^,EntrySize);
  END
  ELSE
    SetData:=FALSE;
  dispose(hfile,done);
END;

FUNCTION TFHASH.Delete(key:keySTRING;var data):BOOLEAN;
VAR p,p2:LONGINT;collcount:byte;
BEGIN
  HFile:=New(PDosStream,Init(dateiname^,stOpen));
  IF HFile^.Status<>stOK THEN
  BEGIN
    dispose(hfile,done);
    HashError:=1;
    exit;
  END;
  collcount:=0;
  p:=HashFunc(KondStr(key));
  p:=FileOffset+(p*EntrySize);
  p2:=p;
  HFile^.Seek(p);
  HFile^.Read(ActEntry^,EntrySize);
  WHILE (ActEntry^.Key<>'') AND (ActEntry^.Key<>key) AND (ActEntry^.next<>0) DO
  BEGIN
    inc(collcount);
    p2:=actEntry^.Next;
    HFile^.Seek(actEntry^.Next);
    HFile^.Read(actEntry^,EntrySize);
  END;
  IF ActEntry^.key=key THEN
  BEGIN
    Delete:=TRUE;
    ActEntry^.Key:='';
    HFile^.Write(ActEntry^,EntrySize);
  END
  ELSE
    Delete:=TRUE;
  dispose(hfile,done);
END;
{$IFDEF FPK}
{$PACKRECORDS NORMAL}
{$ENDIF}
END.