(* MCUT - A Packet-Radio System for DOS and LinuX
 *  unit GNOM - (c) 1997 Hans-Peter Zorn, DG4IAD
 *  This Unit provides a GNOM-driver interface. GNOM-Drivers are
 *  high-level Packet-Radio drivers as used in Jochen Sarrazins GNOM
 *
 *  Currently I support only one GNOM-Driver at once.
 *)


UNIT GNOM;

INTERFACE

VAR GnomError:LONGINT
 
 (* initializes the interface.
  *  sock is the pathname of the socket to the GNOM-Driver
  *  gives back a driver-id or a negative value when the
  *  initialization failed.
  *)
function AttachGnom(sock:STRING):LONGINT;
 
(* Detach the Gnom-Driver,
 *  id is the driver-id (in case multiple drivers are supported)
 *)

procedure DetachGnom(id	: LONGINT);

function PortCount(id: LONGINT):WORD;

procedure GnomMakeConnect(port, source, dest: STRING);

{$PACKRECORD 1} (* We must byte-align the records to be C-Compatible) *)


IMPLEMENTATION

USES strings,sockets,iadvar;

CONST
  MAXDRV=16;
  MAXPORTS=16;
  DriverCount:BYTE=0;

TYPE
  PortInfo=RECORD
	     name,description: STRING;
	     baud: LONGINT;
	   END;		      
		  
  GnomDrvRec=RECORD
	       portsCount: BYTE;
	       ports: ARRAY[1..MAXPORTS] OF PortInfo;
	       DriverName: STRING;
	       fd: LONGINT;
	     END;

  comm_port_info=RECORD
		   port: ARRAY[1..20] OF CHAR;
		   desc: ARRAY[1..50] OF CHAR;
		   baud: LONGINT;

		 END;
  comm_got_connect=RECORD
		     port: ARRAY[1..20] OF CHAR;
		     call: ARRAY[1..10] OF CHAR;
		     path: ARRAY[1..100] OF CHAR;
		     mycall: ARRAY[1..10] OF CHAR;
		     qsonr: LONGINT
		     old: LONGINT;
		   END;
  comm_make_connect=RECORD
		      port: ARRAY[1..20] OF CHAR;
		      mycall: ARRAY[1..10] OF CHAR;
		      call: ARRAY[1..10]OF CHAR;
		      digis: ARRAY[1..100] OF CHAR;
		      chan: LONGINT
		    END;
  comm_listen_to=RECORD
		   port: ARRAY[1..20] OF CHAR;
		   call: ARRAY[1..10] OF CHAR;
		 END;

		 

VAR
  drivers:ARRAY[1..MAXDRV] OF GnomDrvRec;
 
function AttachGnom(path: STRING):LONGINT;
VAR cmd:BYTE;
BEGIN
  IF DriverCount=MAXDRV THEN
    exit;
  inc(DriverCount);
  WITH Drivers[DriverCount] DO
  BEGIN
    fd:=socket(AF_UNIX,SOCK_STREAM,0);
    Servaddr.family:=AF_UNIX;
    StrPCopy(servaddr.path,path);
    servlen:=length(tfkiss_path)+3;
    sockets.connect(fd,ServAddr,servlen,sin,sout);
    if SocketError<>0 THEN
    BEGIN
      sys_close(fd);
      str(SocketError,s);
      popup('connect:'+s);
      AttachGnom:=-1;
      dec(DriverCount);
      exit;
    END;
    (* Ok, we are connected. Get the Driver-Informations... *)
    cmd:=1;
    fcntl(fd,NOT O_NONBLOCK);
    Sys_Write(fd,@cmd,1);
    IF (Sys_Read(fd,@infos,SizeOf(comm_drv_info))<>SizeOf(comm_drv_info)) THEN
    BEGIN
      PopUp('GNOM-protocole error');
      sys_close(fd);
      AttachGnom:=-1;
      exit;
    END;
    DriverName:=StrPas(@infos.name[1]);
    IF sys_read(fd,@PortsCount,1)<>1 THEN
      exit;
    FOR B:= 1 to PortsCount DO
    BEGIN
      IF sys_read(fd,@c_port_info,SizeOF(comm_port_info))<>SizeOF(comm_port_info) THEN
	exit;
      ports[PortsCount].name:=StrPas(@c_port_info.port[1]);
      ports[PortsCount].description:=StrPas(@c_port_info.desc[1]);
      ports[PortsCount].baud:=c_port_info.baud;
    END;
  END;
  AttachGnom:=DriverCount;
END;



function Chan2QsoNr(chan,drv: BYTE):BYTE;
VAR B:BYTE;
BEGIN
  IF KS[chan]=NIL THEN Chan2QsoNr:=0
  ElSE
    Chan2QsoNR:=KS[chan]^.stream;
 
END; { Chan2QsoNr }

function QsoNr2Chan(qsonr,drv: BYTE):BYTE;
VAR B:BYTE;
BEGIN
  QSONr2Chan:=0;
  FOR B:=1 TO 255 DO
   IF KS[B]<>NIL THEN
     IF (KS[B]^.stream=qsonr) AND (KS[B]^.drv=drv) THEN
     BEGIN
       QSONr2Chan:=B;
       exit;
     END;
END; { QsoNr2Chan }

procedure rec(Var Data;size,fd : longint);
BEGIN
   IF sys_read(fd,data,size)<>size THEN;
END; { rec }

procedure snd(cmd: BYTE; var data;size :longint;drv:BYTE;);
BEGIN
  if sys_write(drivers[drv].fd,@cmd,1)<> 1 THEN;
  if size > 0 THEN
    IF sys_write(drivers[drv].fd,data,size)<>size THEN;
END;
procedure gnom_make_connect(C  : BYTE;_call,_mycall,_port:STRING);
VAR d : comm_make_connect;l,r:STRING;
BEGIN
  WITH d DO
  BEGIN
    port:=StrPas(_port);
    mycall:=StrPas(_mycall);
    if not find(' ',_call,l,r) THEN
    BEGIN
      call:=strpas(l);
      digis:=strpas(r);
    END ELSE  
      call:=strpas(_call);
    chan:=c;
  END;  
  snd(CMD_CONN,d,sizeof(d),Port2Drv(_port)]); 
END; { gnom_make_connect }

procedure gnom_disconnect(drv,qsonr:BYTE);
BEGIN
  snd(CMD_DISC,qsonr,sizeof(qsonr),drv); 
END; { gnom_disconnect }
  

procecdure gnom_send_frame(C: BYTE;S:STRING);
TYPE t=RECORD
	 len: LONGINT;
	 data: ARRAY[0..255] of CHAR;
       END;
VAR
  A:t;
BEGIN
  if S='' THEN exit;
  a.len:=byte(S[0]);
  move(S[1],a.data[0],a.len);
  snd(CMD_SEND,a,sizeof(a),KS[C]^.driver);
END; { GnomMakeConnect }

procedure gnom_listen_to(call,port: STRING);
TYPE t=RECORD
	 port: ARRAY[1..20] OF CHAR;
	 call: ARRAY[1..10] OF CHAR;
       END;   
VAR
  A:T;
BEGIN
  a.port:=strpas(port);
  a.call:=strpas(call);
  snd(CMD_LISTEN,a,sizeof(a),port2drv(port));
END; { gnom_listen_to }

procedure gnom_dont_listen_to(call,port: STRING);
TYPE t=RECORD
	 port: ARRAY[1..20] OF CHAR;
	 call: ARRAY[1..10] OF CHAR;
       END;   
VAR
  A:T;
BEGIN
  a.port:=strpas(port);
  a.call:=strpas(call);
  snd(CMD_STOP_LISTEN,a,sizeof(a),port2drv(port));
END; { gnom_dont_listen_to }

procedure gnom_monitor_activate;
BEGIN

END; { gnom_monitor_activate }

procedure gnom_monitor_deactivate;
BEGIN
END; { gnom_monitor_deactivate }


procedure comm_connect_ack(drv: BYTE);
BEGIN
END; { comm_connect_ack }
procedure comm_connect(drv: BYTE);
BEGIN
END; { comm_connect }

procedure comm_status(drv: BYTE);
BEGIN
END; { comm_status }

procedure comm_disconnect(drv: BYTE);
BEGIN
END; { comm_disconnect }

procedure comm_data(drv: BYTE);
BEGIN
END; { comm_data }

procedure comm_monitor(drv: BYTE);
BEGIN
END; { comm_monitor }

procedure comm_xon(drv: BYTE);
BEGIN
END; { comm_xon }

procedure comm_xoff(drv: BYTE);
BEGIN
END; { comm_xoff }

procedure comm_all_data(drv: BYTE);
BEGIN

END; { comm_all_data }


function DoGnom:LONGINT;
VAR B:BYTE;
      Var
    TRec: Time;
    SelectArray: Array[1..5] of LongInt;
    Regs: SysCallRegs;
    FDS: Array[0..31] of LongInt;
    i: LongInt;
	cmd:BYTE;
BEGIN
  FOR B:=1 TO DriverCount DO
  WITH Drivers[B] DO
  BEGIN
    {okok, spaeter kommt da ein highlevel select()-aufruf hin, wenn
     er implementiert ist...}
    {<EXTREMELY_DIRTY_HACK>}
    TRec.Sec := 0;
    TRec.USec := 1000;
    For i := 0 to 31 Do
      FDS[i] := 0;
    FDS[serialfd div (sizeof(longint)*8)] := (1 shl (fd mod
					      (sizeof(longint)*8)));
    SelectArray[1] := fd+1;
    SelectArray[2] := LongInt(@FDS);
    SelectArray[3] := 0;
    SelectArray[4] := 0;
    SelectArray[5] := LongInt(@TRec);
    Regs.ebx := LongInt(@SelectArray);
    IF (SysCall(Syscall_nr_select, regs) <> 0) THEN
    {</EXTREMELY_DIRTY_HACK>}
    (* ok, the driver has something for us, nice *)
    BEGIN
      if sys_read(fd,&cmd,1)<>1 THEN exit;
      CASE cmd OF
	2  : (* Result of a outgoing connect request *)
	    comm_connect_ack(b);
	3  : comm_connect(b);
	4  : comm_status(b);
	5  : comm_disconnect(b);
	6  : comm_data(b);
	7  : comm_all_data(b);
	40 : comm_xon(b);
	41 : comm_xoff(b);
	50 : comm_monitor(b);
      END; { case }
      exit;
    END; 
  END;
END;
   
END.
{$PACKRECORD 2}















